from pathlib import Path
from typing import Callable, List, Union
from tqdm import tqdm


import numpy as np
import salvus.namespace as sn

import parameterization


def test_material_gradient(
    *,
    all_h: np.ndarray,
    model: Union[Path, str, sn.UnstructuredMesh],
    gradient: Union[Path, str, sn.UnstructuredMesh],
    parameter: str,
    simulation: sn.simple_config.simulation.Waveform,
    m0: float,
    misfit_function: Callable,
    site_name: str,
    ranks: int = 1,
    random: bool = False,
    quiet: bool = False,
) -> List[float]:
    # First, read in the model and gradient.
    if not type(model) == sn.UnstructuredMesh:
        model = sn.UnstructuredMesh.from_h5(str(model))
    if not type(gradient) == sn.UnstructuredMesh:
        gradient = sn.UnstructuredMesh.from_h5(str(gradient))

    # Check sanity of test
    v_name = "Valence"
    mm_name = "FemMassMatrix"
    gradient_pars_all = set(gradient.elemental_fields.keys())
    gradient_pars_base = set(x.split("_")[0] for x in gradient_pars_all)
    gradient_pars = set(
        p for p in gradient_pars_all if mm_name not in p if v_name not in p
    )
    assert parameter in gradient_pars_base, parameter
    for g_par in gradient_pars:
        if g_par == "TTISymmetryAxis":
            continue
        if g_par == "RotationConvention":
            continue
        assert g_par in model.elemental_fields.keys(), g_par

    field = model.elemental_fields[f"{parameter}"]
    if random:
        dm = 0.01 * field * np.random.random() + field.mean()
    else:
        dm = field

    output = []
    for h in tqdm(all_h, desc=f"Gradient test ({parameter})", disable=quiet):
        # We are doing a relative perturbation, so copy the original model.
        p_model = model.copy()
        p_model.tensorized = True

        # Perturb all parameters by the same amount.
        field = p_model.elemental_fields[f"{parameter}"]
        field += h * dm

        simulation.set_mesh(p_model)
        p_model.write_h5("test_model.h5")
        sn.api.run(
            ranks=ranks,
            input_file=simulation,
            get_all=True,
            site_name=site_name,
            output_folder="gradient_test",
            delete_remote_files=False,
            overwrite=True,
            verbosity=0,
        )

        g_fields = gradient.elemental_fields
        rhs = np.sum(
            (
                g_fields[f"{parameter}"]
                * g_fields[f"{mm_name}"]
                / g_fields[f"{v_name}"]
            )
            * dm
        )

        new_misfit = misfit_function("gradient_test/receivers.h5")
        lhs = (new_misfit - m0) / h
        rel_error = (lhs - rhs) / np.abs(rhs)
        output.append(np.abs(rel_error))

    return output


def get_unit_cube_mesh(
    *,
    dim: int,
    model_order: int,
    n_elem_per_dim: int,
    par: parameterization.AbstractParameterBase,
    deform: bool = True,
) -> sn.UnstructuredMesh:
    """
    A simple factory function to return a unit hyperrcube mesh with an
    appropriate set of parameters.

    Args:
        dim: Dimension of the mesh.
        model_order: Order of Lagrange polynomials used to interpolate the
            model.
        n_elem_per_dim:  Number of elements per dimension of the hypercube.
        param par: Physical parameterization of the model.
    """

    # Sanity checks.
    assert dim in (2, 3)

    # Get underlying mesh object
    if dim == 2:
        mesh = sn.simple_mesh.basic_mesh.CartesianHomogeneousAcoustic2D(
            x_max=1.0,
            y_max=1.0,
            vp=1.0,
            rho=1.0,
            tensor_order=model_order,
            max_frequency=float(n_elem_per_dim),
            elements_per_wavelength=1.0,
        ).create_mesh()
    else:
        mesh = sn.simple_mesh.basic_mesh.CartesianHomogeneousAcoustic3D(
            x_max=1.0,
            y_max=1.0,
            z_max=1.0,
            vp=1.0,
            rho=1.0,
            tensor_order=model_order,
            max_frequency=float(n_elem_per_dim),
            elements_per_wavelength=1.0,
        ).create_mesh()

    # Attach parameters.
    mesh.elemental_fields = {}
    is_acoustic = par.physics() == parameterization.Physics.Acoustic
    mesh.attach_field("fluid", np.ones(mesh.nelem) * is_acoustic)

    for name, val in par.parameters():
        if name != "AXIS":
            mesh.attach_field(
                name, np.ones_like(mesh.get_element_nodes()[:, :, 0]) * val
            )
        else:
            for i, a in enumerate(val):
                mesh.attach_global_variable(f"a_{i}", a)

    # Deform
    if deform and dim == 2:
        amp = 0.05
        k = 2 * np.pi
        px, py = mesh.points[:, 0], mesh.points[:, 1]
        mesh.points[:, 0] += amp * np.sin(k * px) * np.sin(k * py)
        mesh.points[:, 1] += amp * np.sin(k * px) * np.sin(k * py)

    if deform and dim == 3:
        amp = 0.05
        k = 2 * np.pi
        px, py, pz = mesh.points[:, 0], mesh.points[:, 1], mesh.points[:, 2]
        mesh.points[:, 0] += amp * np.sin(k * px) * np.sin(k * py)
        mesh.points[:, 1] += amp * np.sin(k * px) * np.sin(k * py)
        mesh.points[:, 2] += amp * np.sin(k * pz) * np.sin(k * px)

    return mesh
