# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# This file is part of the Salvus Software Suite.
#
# Copyright (c) 2018-2023 Mondaic Ltd.
#
# All rights reserved.
#
# NOTICE:  All information contained herein is, and remains
# the property of Mondaic. The intellectual and technical concepts
# contained herein are proprietary to Mondaic and are protected by trade
# secret or copyright law, and may be covered by European, US and/or
# Foreign Patents, or patents in process.
# Dissemination of this information or reproduction of this material
# in any form is strictly forbidden unless prior written permission is
# obtained from Mondaic.
#
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import abc
from enum import Enum
from typing import Union, Optional

import numpy as np


class Physics(Enum):
    Acoustic = 1
    Elastic = 2


class AcousticParameterizations(Enum):
    Compliance = 1
    RhoVp = 2


class ElasticParameterizations(Enum):
    Compliance = 1
    RhoVpVs = 2
    FullyAnistropic3D = 3
    Tti3D = 3


class AbstractParameterBase(metaclass=abc.ABCMeta):
    @abc.abstractstaticmethod
    def p(self):
        raise NotImplementedError

    @abc.abstractstaticmethod
    def physics(self):
        raise NotImplementedError

    @abc.abstractclassmethod
    def from_compliance(cls, compliance: "AbstractParameterBase"):
        raise NotImplementedError

    @abc.abstractmethod
    def to_compliance(self):
        raise NotImplementedError

    @abc.abstractproperty
    def json_string(self):
        raise NotImplementedError

    def convert(self, parameterization):
        if self.physics() == Physics.Acoustic:
            if parameterization.p() == AcousticParameterizations.Compliance:
                return self.to_compliance()
            elif parameterization.p() == AcousticParameterizations.RhoVp:
                return AcousticRhoVp.from_compliance(self.to_compliance())
            else:
                raise ValueError(
                    "Cannot transform an acoustic parameterization to an "
                    "elastic parameterization."
                )

        elif self.physics() == Physics.Elastic:
            if parameterization.p() == ElasticParameterizations.Compliance:
                return self.to_compliance()
            elif parameterization.p() == ElasticParameterizations.RhoVpVs:
                return ElasticRhoVpVs.from_compliance(self.to_compliance())

        else:
            raise NotImplementedError(
                "Only acoustic and elastic physics currently implemented."
            )

    def parameters(self):
        for key, val in self.__dict__.items():
            yield key.split("_")[1].upper(), val


class AcousticCompliance(AbstractParameterBase):
    def __init__(
        self, *, m0: Union[float, np.ndarray], m1: Union[float, np.ndarray]
    ) -> None:
        self._m0 = m0
        self._m1 = m1

    @staticmethod
    def p() -> AcousticParameterizations:  # type: ignore
        return AcousticParameterizations.Compliance

    @staticmethod
    def physics() -> Physics:  # type: ignore
        return Physics.Acoustic

    @classmethod
    def from_compliance(  # type: ignore
        cls, compliance: AbstractParameterBase
    ):
        return compliance

    @property
    def json_string(self):
        return "linear"

    def to_compliance(self):
        return self


class AcousticRhoVp(AbstractParameterBase):
    def __init__(
        self, *, rho: Union[float, np.ndarray], vp: Union[float, np.ndarray]
    ) -> None:
        self._vp = vp
        self._rho = rho

    @staticmethod
    def p() -> AcousticParameterizations:  # type: ignore
        return AcousticParameterizations.RhoVp

    @staticmethod
    def physics() -> Physics:  # type: ignore
        return Physics.Acoustic

    @classmethod
    def from_compliance(cls, compliance: AcousticCompliance):  # type: ignore
        rho = 1 / compliance._m1
        vp = np.sqrt(compliance._m1 / compliance._m0)
        return cls(rho=rho, vp=vp)

    @property
    def json_string(self):
        return "rho-vp"

    def to_compliance(self) -> AcousticCompliance:
        m1 = 1 / self._rho
        m0 = m1 / self._vp**2
        return AcousticCompliance(m0=m0, m1=m1)


class ElasticCompliance(AbstractParameterBase):
    def __init__(
        self,
        *,
        lam: Union[float, np.ndarray],
        mu: Union[float, np.ndarray],
        rho: Union[float, np.ndarray],
    ) -> None:
        self._mu = mu
        self._rho = rho
        self._lambda = lam

    @staticmethod
    def p() -> ElasticParameterizations:  # type: ignore
        return ElasticParameterizations.Compliance

    @staticmethod
    def physics() -> Physics:  # type: ignore
        return Physics.Elastic

    @classmethod
    def from_compliance(  # type: ignore
        cls, compliance: AbstractParameterBase
    ):
        return compliance

    @property
    def json_string(self):
        return "linear"

    def to_compliance(self):
        return self


class ElasticTti3D(AbstractParameterBase):
    def __init__(
        self,
        *,
        rho: Union[float, np.ndarray],
        vph: Union[float, np.ndarray],
        vpv: Union[float, np.ndarray],
        vsh: Union[float, np.ndarray],
        vsv: Union[float, np.ndarray],
        eta: Union[float, np.ndarray],
        axis: np.ndarray = np.array([0.0, 0.0, 0.0]),
    ) -> None:
        self._rho = rho
        self._vph = vph
        self._vpv = vpv
        self._vsh = vsh
        self._vsv = vsv
        self._eta = eta

        if len(axis) != 3:
            raise ValueError(
                "Need a 3-dimensional vector for the symmetry axis!"
            )

        self._axis = axis

    @staticmethod
    def p() -> ElasticParameterizations:  # type: ignore
        return ElasticParameterizations.Tti3D

    @staticmethod
    def physics() -> Physics:  # type: ignore
        return Physics.Elastic

    @classmethod
    def from_compliance(cls, compliance: ElasticCompliance):  # type: ignore
        c = compliance
        vsv = np.sqrt(c._mu / c._rho)
        vpv = np.sqrt((c._lambda + 2 * c._mu) / c._rho)

        vph = vpv
        vsh = vsv
        rho = c._rho
        eta = 1.0

        return cls(vph=vph, vsh=vsh, vpv=vpv, vsv=vsv, rho=rho, eta=eta)

    def json_string(self):
        return "tti"

    def to_compliance(self):
        vp = (2 / 3) * self._vph + (1 / 3) * self._vpv
        vs = (1 / 3) * self._vsh + (2 / 3) * self._vsv
        rho = self._rho

        m = rho * vs**2
        lam = rho * (vp**2 - 2 * vs**2)
        return ElasticCompliance(lam=lam, mu=m, rho=rho)


class ElasticFullyAnisotropic3D(AbstractParameterBase):
    def __init__(
        self,
        *,
        rho: Union[float, np.ndarray],
        c11: Union[float, np.ndarray],
        c12: Union[float, np.ndarray],
        c13: Union[float, np.ndarray],
        c14: Optional[Union[float, np.ndarray]] = None,
        c15: Optional[Union[float, np.ndarray]] = None,
        c16: Optional[Union[float, np.ndarray]] = None,
        c22: Union[float, np.ndarray],
        c23: Union[float, np.ndarray],
        c24: Optional[Union[float, np.ndarray]] = None,
        c25: Optional[Union[float, np.ndarray]] = None,
        c26: Optional[Union[float, np.ndarray]] = None,
        c33: Union[float, np.ndarray],
        c34: Optional[Union[float, np.ndarray]] = None,
        c35: Optional[Union[float, np.ndarray]] = None,
        c36: Optional[Union[float, np.ndarray]] = None,
        c44: Union[float, np.ndarray],
        c45: Optional[Union[float, np.ndarray]] = None,
        c46: Optional[Union[float, np.ndarray]] = None,
        c55: Union[float, np.ndarray],
        c56: Optional[Union[float, np.ndarray]] = None,
        c66: Union[float, np.ndarray],
    ) -> None:
        self._rho = rho
        self._c11 = c11
        self._c12 = c12
        self._c13 = c13
        self._c14 = c14 if c14 else 0 * c11
        self._c15 = c15 if c15 else 0 * c11
        self._c16 = c16 if c16 else 0 * c11
        self._c22 = c22
        self._c23 = c23
        self._c24 = c24 if c24 else 0 * c11
        self._c25 = c25 if c25 else 0 * c11
        self._c26 = c26 if c26 else 0 * c11
        self._c33 = c33
        self._c34 = c34 if c34 else 0 * c11
        self._c35 = c35 if c35 else 0 * c11
        self._c36 = c36 if c36 else 0 * c11
        self._c44 = c44
        self._c45 = c45 if c45 else 0 * c11
        self._c46 = c46 if c46 else 0 * c11
        self._c55 = c55
        self._c56 = c56 if c56 else 0 * c11
        self._c66 = c66
        self._axis = np.zeros(3)

    @staticmethod
    def p() -> ElasticParameterizations:  # type: ignore
        return ElasticParameterizations.FullyAnistropic3D

    @staticmethod
    def physics() -> Physics:  # type: ignore
        return Physics.Elastic

    @classmethod
    def from_compliance(cls, compliance: ElasticCompliance):  # type: ignore
        c11 = compliance._lambda + 2 * compliance._mu
        c22 = compliance._lambda + 2 * compliance._mu
        c33 = compliance._lambda + 2 * compliance._mu

        c44 = compliance._mu
        c55 = compliance._mu
        c66 = compliance._mu

        c12 = compliance._lambda
        c13 = compliance._lambda
        c23 = compliance._lambda

        return cls(
            c11=c11,
            c12=c12,
            c13=c13,
            c22=c22,
            c23=c23,
            c33=c33,
            c44=c44,
            c55=c55,
            c66=c66,
            rho=compliance._rho,
        )

    def json_string(self):
        return "fully-anisotropic"

    def to_compliance(self):
        lam = (
            6 * self._c11
            + self._c33
            + 8 * self._c13
            - 4 * self._c44
            - 10 * self._c66
        ) / 15

        mu = (
            self._c11
            + self._c33
            - 2 * self._c13
            + 6 * self._c44
            + 5 * self._c66
        ) / 15

        return ElasticCompliance(lam=lam, mu=mu, rho=self._rho)


class ElasticRhoVpVs(AbstractParameterBase):
    def __init__(
        self,
        *,
        rho: Union[float, np.ndarray],
        vp: Union[float, np.ndarray],
        vs: Union[float, np.ndarray],
    ) -> None:
        self._vp = vp
        self._vs = vs
        self._rho = rho

    @staticmethod
    def p() -> ElasticParameterizations:  # type: ignore
        return ElasticParameterizations.RhoVpVs

    @staticmethod
    def physics() -> Physics:  # type: ignore
        return Physics.Elastic

    @classmethod
    def from_compliance(cls, compliance: ElasticCompliance):  # type: ignore
        vp = np.sqrt(
            (compliance._lambda + 2 * compliance._mu) / compliance._rho
        )
        vs = np.sqrt(compliance._mu / compliance._rho)
        return cls(rho=compliance._rho, vp=vp, vs=vs)

    @property
    def json_string(self):
        return "rho-vp-vs"

    def to_compliance(self):
        lam = self._rho * (self._vp**2 - 2 * self._vs**2)
        mu = self._rho * self._vs**2
        return ElasticCompliance(lam=lam, mu=mu, rho=self._rho)
